# Develop a Flink application

> **Objectives**: Develop a simple Flink application.
>
> **Complexity**: Beginner
>
> **Time to complete**: 30 min
>
> **Source code**: <a href="develop-flink-application.zip" download target="_blank">Download</a>


This tutorial demonstrates how to develop, debug, and run a simple Flink application that reads data from the stream
layer, and logs all this data using log4j.

The tutorial covers the following topics:

- [Set up the Maven project](#set-up-the-maven-project)
- [Write the source code](#write-the-source-code)
- [Run the application](#run-the-application)
- [Attach the debugger](#attach-the-debugger)
- [Project generation using Maven archetype](#project-generation-using-maven-archetype)

### Set up the Maven project

Download the source code at the beginning of the tutorial and put it in a folder of your choice, or create a folder
structure for your project from scratch:

```txt
develop-flink-application
└── src
    └── main
        ├── java
        └── resources
        └── scala
```

You can do this with a single `bash` command:

```bash
mkdir -p develop-flink-application/src/main/{java,resources,scala}
```

The Maven POM file is similar to the one in the
[Verify Maven Settings](../verify-m2/README.md) example, however, with an updated parent POM and dependencies
sections:

The Parent POM is `sdk-stream-bom_${scala.compat.version}`, because we need to use Flink-related libraries.

```
[snippet](pom.xml#parent)
```

The following dependencies are used:

- `com.here.platform.data.client:local-support_${scala.compat.version}` to read data from a local data catalog.
- `org.apache.flink:flink-clients_${scala.compat.version}` to provide `ExecutorFactory` for the Flink application.
- `org.apache.flink:flink-streaming-java_${scala.compat.version}` to run a Java Flink application.
- `org.apache.flink:flink-streaming-scala_${scala.compat.version}` to run a Scala Flink application.
- `com.here.platform.data.client:flink-support_${scala.compat.version}` to read data from the data catalogs on the
  platform.
- `org.slf4j:slf4j-log4j12` to log application results to the console and to the Splunk on the platform.
- `com.here.platform.pipeline:pipeline-interface_${scala.compat.version}` to get information about input catalogs from the `PipelineContext`.


Dependencies:

```
[snippet](pom.xml#dependencies)
```

Once you have added all the necessary dependencies to the `pom.xml` file, the next step would be to write the code of
the application and run it.

## Write the source code

As mentioned before, the tutorial shows how to write a simple Flink application that reads data from a stream layer and
outputs the data in the console. The data is added to the stream layer during the execution of your stream application
using OLP CLI. All data read from the layer is logged to the console at the same time
using [log4j](https://logging.apache.org/log4j/2.x/). The configuration of log4j is located in
the [`src/main/resources/log4j.properties`](src/main/resources/log4j.properties) file.

Let's look at the implementation of this Flink application. In the code snippet below, you can see that
the `FlinkQueryApi` is used to create a subscription to receive new data added to the stream layer. Once the subscription is
created and the `DataStreamSource` source function that emits new data from stream layer is declared, we map partitions
using the `SampleMaper` class that extends `RichMapFunction`, which provides setup and teardown methods.
The `SampleMapper` class downloads partitions and logs as a human-readable string.

{% codetabs name="Java", type="java" -%} {% raw %}
[snippet](src/main/java/DevelopFlinkApplication.java)
{% endraw %} {% language name="Scala", type="scala" -%} {% raw %}
[snippet](src/main/scala/DevelopFlinkApplicationScala.scala)
{% endraw %} {%- endcodetabs %}

Once the code is written, you can prepare resources and run the application.

## Run the application

To run the application, you need to prepare the resources - create a catalog with a stream layer where
the data will be written.

In this tutorial we will run our application locally, therefore, it will be enough for us to create a local catalog. No
authentication and no access to the external network are needed to run this tutorial as we use local catalogs. Since
they are contained in a local machine, they are not subject to naming conflicts within your realm and you can use any
name you want.

To create a local catalog with a stream layer, you will need the `catalog-with-stream-layer.json` config file that you downloaded together with the source code at the beginning of the tutorial. It contains configuration allowing you to create a catalog with a layer using only one OLP CLI command:

```bash
olp local catalog create streaming-catalog streaming-catalog --config catalog-with-stream-layer.json
```

The structure of the `catalog-with-stream-layer.json` file is as follows:
{% codesnippet language="json" %}
[snippet](catalog-with-stream-layer.json)
{% endcodesnippet %}

> #### Note
> If a billing tag is required in your realm, update the config file by adding the `billingTags: ["YOUR_BILLING_TAG"]` property to the `layer` section.

As mentioned in the [Set up the Maven project](#set-up-the-maven-project) chapter, the `PipelineContext` is used
to get information about the input catalog from the `pipeline-config.conf` file.
The structure of the `pipeline-config.conf` file is as follows:

{% codesnippet language="conf" %}
[snippet](pipeline-config.conf)
{% endcodesnippet %}

Although we do not use the output catalog in our tutorial,
we can use the input catalog HRN to replace both `OUTPUT_CATALOG_HRN` and `INPUT_CATALOG_HRN` in the `pipeline-config.conf` file.


After you have replaced the placeholder, run the application from the root of the downloaded tutorial using the following
command:

{% codetabs name="Java", type="sh" -%}
{% raw %}
mvn compile exec:java -D"exec.mainClass"="DevelopFlinkApplication" \
-Dhere.platform.data-client.endpoint-locator.discovery-service-env=local \
-Dspark.master=local[*] \
-Dpipeline-config.file=pipeline-config.conf
{% endraw %}
{% language name="Scala", type="sh" -%}
{% raw %}
mvn compile exec:java -D"exec.mainClass"="DevelopFlinkApplicationScala"
-Dhere.platform.data-client.endpoint-locator.discovery-service-env=local \
-Dspark.master=local[*] \
-Dpipeline-config.file=pipeline-config.conf
{% endraw %}
{%- endcodetabs %}

The command has the following parameters:

- `exec.mainClass` – entry point to run your application.
- `here.platform.data-client.endpoint-locator.discovery-service-env=local` – configures the Data Client Library to only
  use local catalogs.

At this moment, the application is running, but there is no data in the input catalog. Let's put some partitions in it. Open a new
console window and run the following bash script that uploads data with the `First HERE Platform Flink Application`
content to the stream layer 10 times every 5 seconds:

{% codetabs name="Bash", type="bash" -%} {% raw %}
[snippet](scripts/populate-streaming-data.sh)
{% endraw %} {%- endcodetabs %}

To run the script, execute the following command:

```bash
bash scripts/populate-streaming-data.sh {{YOUR_CATALOG_HRN}}
```

Once the script is running, you should see the logs with the same information in the console of the running application.

## Attach the debugger

In this chapter you will learn how to debug Flink applications using the capabilities of Intellij Idea, or rather you
will learn how to [attach to the process](https://www.jetbrains.com/help/idea/attaching-to-local-process.html)
to start debugging if you run your program through the console.

In order to configure the debugger, you need to set up the `MAVEN_OPTS` variable. To do so, stop the running application
and run the following command:

{% codetabs name="JDK 5-8", type="sh" -%} 
{% raw %} 
export MAVEN_OPTS=-agentlib:jdwp=transport=dt_socket,server=y,suspend=n,address=5005 
{% endraw %}
{% language name="Java 9 or later", type="sh" -%}
{% raw %} 
export MAVEN_OPTS=-agentlib:jdwp=transport=dt_socket,server=y,suspend=n,address=*:5005 
{% endraw %} 
{%- endcodetabs %}

The option has the following parameters:

- `address` – the port that will be used for debugging. The tutorial uses port `5005`, however, you can use any
  free port.
- `server=y` – specifies that the process should listen for incoming debugger connections (act as a server).
- `suspend=n` – specifies that the process should wait until the debugger has been connected.

Now you can run your application using the following command:

{% codetabs name="Java", type="sh" -%} {% raw %} mvn compile exec:java -D"exec.mainClass"="DevelopFlinkApplication"
-Dhere.platform.data-client.endpoint-locator.discovery-service-env=local {% endraw %} {% language name="Scala", type="
sh" -%} {% raw %} mvn compile exec:java -D"exec.mainClass"="DevelopFlinkApplicationScala"
-Dhere.platform.data-client.endpoint-locator.discovery-service-env=local {% endraw %} {%- endcodetabs %}

Make sure that line `Listening for transport dt_socket at address: 5005` appears in the logs.

Let's set the breakpoint before we attach to the running application, for example, in the line where the partitions are mapped.
From this line, you can get a lot of useful information about the partition you downloaded:

![](img/checkpoint.png)

Now we can attach to the process using `Run` > `Attach to Process` and select the process with the specified `5050`
port.

![](img/atach-to-process.png)

Once the process is attached and your program is running and waiting for the data to appear in the stream layer, you
need to run the script again to inject data:

```bash
bash scripts/populate-streaming-data.sh {{YOUR_CATALOG_HRN}}
```

After running the `scripts/populate-streaming-data.sh` script, the debugger should stop at the breakpoint as soon as new
data appears in the catalog and your application will start reading it. Now you can step through the code and inspect
the content of the variables and the stack trace.

![](img/debug.png)

You can also use the
standard [Java Debugger](https://docs.oracle.com/javase/7/docs/technotes/tools/windows/jdb.html#:~:text=The%20Java%20Debugger%2C%20jdb%2C%20is,or%20remote%20Java%20Virtual%20Machine.)
instead of the debugger in the Intellij IDEA.

## Project generation using a Maven archetype
You can use a Maven archetype to bootstrap a Maven project for a Flink application. In this case, the project is set up faster with the following tasks completed automatically:
* Inclusion of the SDK BOM file.
* Creation of the Maven profile that generates the fat JAR for the platform.

For more information on your streaming pipeline options, see [SDK Workflows]({{book.filtered.devGuidesUrlBase}}/java-scala-dev/dev_guide/topics/streaming-pipeline.html).

To create a Flink application project, use the following command for the Java project:

{% codetabs name="Linux", type="bash" -%}
mvn archetype:generate -DarchetypeGroupId=com.here.platform \
-DarchetypeArtifactId=streaming-java-archetype \
-DarchetypeVersion={{book.filtered.dplArchetypesVersion}} \
-DgroupId=com.here.platform.tutorial \
-DartifactId=develop-flink-application \
-Dversion=1.0-SNAPSHOT \
-Dpackage=com.here.platform.tutorial
{%- language name="Windows", type="bat" -%}
mvn archetype:generate -DarchetypeGroupId=com.here.platform ^
-DarchetypeArtifactId=streaming-java-archetype ^
-DarchetypeVersion={{book.filtered.dplArchetypesVersion}} ^
-DgroupId=com.here.platform.tutorial ^
-DartifactId=develop-flink-application ^
-Dversion=1.0-SNAPSHOT ^
-Dpackage=com.here.platform.tutorial
{%- endcodetabs %}

For the Scala project use the following command:

{% codetabs name="Linux", type="bash" -%}
mvn archetype:generate -DarchetypeGroupId=com.here.platform \
-DarchetypeArtifactId=streaming-scala-archetype \
-DarchetypeVersion={{book.filtered.dplArchetypesVersion}} \
-DgroupId=com.here.platform.tutorial.scala \
-DartifactId=develop-flink-application \
-Dversion=1.0-SNAPSHOT \
-Dpackage=com.here.platform.tutorial
{%- language name="Windows", type="bat" -%}
mvn archetype:generate -DarchetypeGroupId=com.here.platform ^
-DarchetypeArtifactId=streaming-scala-archetype ^
-DarchetypeVersion={{book.filtered.dplArchetypesVersion}} ^
-DgroupId=com.here.platform.tutorial.scala ^
-DartifactId=develop-flink-application ^
-Dversion=1.0-SNAPSHOT ^
-Dpackage=com.here.platform.tutorial
{%- endcodetabs %}


### Build your project to run locally

To build your project, run the following command in your project folder.

```bash
mvn install
```

### Build your project to run on the platform

To run your pipeline on the platform, you need to build a `fat jar` first. To build it, use the following command.

```bash
mvn install -Pplatform
```

For more information on building a `fat jar`, see the [Include the SDK in your project]({{book.filtered.devGuidesUrlBase}}/java-scala-dev/dev_guide/topics/dependency-management.html#include-boms).

## Conclusion

In this tutorial, you have learned the stages of the Flink program development. To learn how to run a Flink application
on the platform and to get acquainted with such monitoring tools as Splunk, Grafana, Flink UI, and Platform Billing
Page, see the [Run a Flink application on the platform](../run-flink-application-platform/README.md) tutorial.

## Further information

For more details on the topics covered in this tutorial, see the following sources:

- For more information on how to develop Flink applications, see the
  [Apache Flink Documentation](https://nightlies.apache.org/flink/flink-docs-stable/).
- For more information on how to work with local catalogs using OLP CLI, see the
  [OLP CLI Guide]({{book.filtered.devGuidesUrlBase}}/open-location-platform-cli/user_guide/topics/local-data/local-catalog-commands.html).
- For more information on how to attach to process to debug your application, see the
  [Debugging code Developer Guide](https://www.jetbrains.com/help/idea/attaching-to-local-process.html)
